@@ -56,6 +56,10 @@
 #include <sstream>
 #include <string>
 
+#ifndef _FOR_R
+#define FOR_R
+#endif
+
 /* This is the package's header */
 #include "isotree.h"
 
@@ -157,7 +161,7 @@ double* set_R_nan_as_C_nan(double *x, size_t n, std::vector<double> &v, int nthr
 {
     v.assign(x, x + n);
     #pragma omp parallel for schedule(static) num_threads(nthreads) shared(x, v, n)
-    for (size_t_for i = 0; i < n; i++)
+    for (size_t_for i = 0; i < (decltype(i))n; i++)
         if (isnan(v[i]))
             v[i] = NAN;
     return v.data();
@@ -166,7 +170,7 @@ double* set_R_nan_as_C_nan(double *x, size_t n, std::vector<double> &v, int nthr
 double* set_R_nan_as_C_nan(double *x, size_t n, int nthreads)
 {
     #pragma omp parallel for schedule(static) num_threads(nthreads) shared(x, n)
-    for (size_t_for i = 0; i < n; i++)
+    for (size_t_for i = 0; i < (decltype(i))n; i++)
         if (isnan(x[i]))
             x[i] = NAN;
     return x;
@@ -532,13 +536,8 @@ Rcpp::RawVector fit_tree(SEXP model_R_ptr,
     else
         ext_model_ptr  =  static_cast<ExtIsoForest*>(R_ExternalPtrAddr(model_R_ptr));
 
-    std::vector<ImputeNode>  *imp_ptr = NULL;
     if (build_imputer)
-    {
         imputer_ptr = static_cast<Imputer*>(R_ExternalPtrAddr(imp_R_ptr));
-        imputer_ptr->imputer_tree.emplace_back();
-        imp_ptr     = &imputer_ptr->imputer_tree.back();
-    }
 
     add_tree(model_ptr, ext_model_ptr,
              numeric_data_ptr,  ncols_numeric,
@@ -554,7 +553,7 @@ Rcpp::RawVector fit_tree(SEXP model_R_ptr,
              min_gain, missing_action_C,
              cat_split_type_C, new_cat_action_C,
              depth_imp_C, weigh_imp_rows_C, all_perm,
-             imp_ptr, min_imp_obs, (uint64_t)random_seed);
+             imputer_ptr, min_imp_obs, (uint64_t)random_seed);
 
     if (ndim == 1)
         return serialize_cpp_obj(model_ptr);
@@ -1013,7 +1012,7 @@ Rcpp::IntegerMatrix csc_to_dense_int
     int *restrict out = INTEGER(out_);
     for (size_t col = 0; col < ncols; col++)
     {
-        for (size_t ix = Xc_indptr[col]; ix < Xc_indptr[col+1]; ix++)
+        for (auto ix = Xc_indptr[col]; ix < Xc_indptr[col+1]; ix++)
             out[(size_t)Xc_ind[ix] + col*nrows]
                 =
             (Xc[ix] >= 0 && !ISNAN(Xc[ix]))?
@@ -1028,7 +1027,7 @@ Rcpp::IntegerMatrix csr_to_dense_int
     real_vec Xr,
     int_vec Xr_ind,
     int_vec Xr_indptr,
-    size_t ncols
+    int ncols
 )
 {
     size_t nrows = Xr_indptr.size() - 1;
@@ -1036,7 +1035,7 @@ Rcpp::IntegerMatrix csr_to_dense_int
     int *restrict out = INTEGER(out_);
     for (size_t row = 0; row < nrows; row++)
     {
-        for (size_t ix = Xr_indptr[row]; ix < Xr_indptr[row+1]; ix++)
+        for (auto ix = Xr_indptr[row]; ix < Xr_indptr[row+1]; ix++)
             out[row + (size_t)Xr_ind[ix]*nrows]
                 =
             (Xr[ix] >= 0 && !ISNAN(Xr[ix]))?
@@ -1051,7 +1050,7 @@ Rcpp::List call_take_cols_by_slice_csr
     Rcpp::NumericVector Xr_,
     Rcpp::IntegerVector Xr_ind_,
     Rcpp::IntegerVector Xr_indptr,
-    size_t ncols_take,
+    int ncols_take,
     bool as_dense
 )
 {
@@ -1064,7 +1063,7 @@ Rcpp::List call_take_cols_by_slice_csr
     size_t total_size = 0;
     for (size_t row = 0; row < nrows; row++)
     {
-        for (size_t col = Xr_indptr[row]; col < Xr_indptr[row+1]; col++)
+        for (auto col = Xr_indptr[row]; col < Xr_indptr[row+1]; col++)
             total_size += Xr_ind[col] < ncols_take;
         out_Xr_indptr[row+1] = total_size;
     }
@@ -1114,8 +1113,8 @@ Rcpp::List call_take_cols_by_index_csr
 )
 {
     /* 'cols_take' should be sorted */
-    size_t n_take = cols_take.size();
-    size_t nrows = Xr_indptr.size() - 1;
+    int n_take = cols_take.size();
+    int nrows = Xr_indptr.size() - 1;
     std::vector<double> out_Xr;
     std::vector<int> out_Xr_ind;
     std::vector<int> out_Xr_indptr(nrows + 1);
@@ -1125,10 +1124,10 @@ Rcpp::List call_take_cols_by_index_csr
     int *restrict ptr_Xr_ind = INTEGER(Xr_ind);
     int *restrict ptr_cols_take = INTEGER(cols_take);
     int *restrict ptr_cols_take_end = ptr_cols_take + n_take;
-    size_t curr_col;
+    int curr_col;
     int *search_res;
 
-    for (size_t row = 0; row < nrows; row++)
+    for (int row = 0; row < nrows; row++)
     {
         curr_ptr = ptr_Xr_ind + Xr_indptr[row];
         end_ptr = ptr_Xr_ind + Xr_indptr[row+1];
@@ -1468,4 +1467,10 @@ void inplace_add(SEXP add_to, SEXP add_this)
     *val = *val + *summand;
 }
 
+// [[Rcpp::export(rng = false)]]
+SEXP deepcopy_int(SEXP x)
+{
+    return Rf_ScalarInteger(Rf_asInteger(x));
+}
+
 #endif /* _FOR_R */
