@@ -67,20 +67,18 @@ void split_itree_recursive(std::vector<IsoTree>     &trees,
                           model_params.min_imp_obs);
     }
 
-    /* check for potential isolated leafs */
-    if (workspace.end == workspace.st || curr_depth >= model_params.max_depth)
-        goto terminal_statistics;
-
-    /* with 2 observations and no weights, there's only 1 potential or assumed split */
-    if ((workspace.end - workspace.st) == 1 && !workspace.weights_arr.size() && !workspace.weights_map.size())
+    /* check for potential isolated leafs or unique splits */
+    if (workspace.end == workspace.st || (workspace.end - workspace.st) == 1 || curr_depth >= model_params.max_depth)
         goto terminal_statistics;
 
     /* when using weights, the split should stop when the sum of weights is <= 1 */
-    sum_weight = calculate_sum_weights(workspace.ix_arr, workspace.st, workspace.end, curr_depth,
-                                       workspace.weights_arr, workspace.weights_map);
-
-    if (curr_depth > 0 && (workspace.weights_arr.size() || workspace.weights_map.size()) && sum_weight <= 1)
-        goto terminal_statistics;
+    if (workspace.changed_weights)
+    {
+        sum_weight = calculate_sum_weights(workspace.ix_arr, workspace.st, workspace.end, curr_depth,
+                                           workspace.weights_arr, workspace.weights_map);
+        if (curr_depth > 0 && sum_weight <= 1)
+            goto terminal_statistics;
+    }
 
     /* for sparse matrices, need to sort the indices */
     if (input_data.Xc_indptr != NULL && impute_nodes == NULL)
@@ -119,7 +117,7 @@ void split_itree_recursive(std::vector<IsoTree>     &trees,
             {
                 if (input_data.Xc_indptr == NULL)
                 {
-                    if (!workspace.weights_arr.size() && !workspace.weights_map.size())
+                    if (!workspace.changed_weights)
                         workspace.this_gain = eval_guided_crit(workspace.ix_arr.data(), workspace.st, workspace.end,
                                                                input_data.numeric_data + workspace.col_chosen * input_data.nrows,
                                                                workspace.buffer_dbl.data(), false,
@@ -149,7 +147,7 @@ void split_itree_recursive(std::vector<IsoTree>     &trees,
 
                 else
                 {
-                    if (!workspace.weights_arr.size() && !workspace.weights_map.size())
+                    if (!workspace.changed_weights)
                         workspace.this_gain = eval_guided_crit(workspace.ix_arr.data(), workspace.st, workspace.end,
                                                                workspace.col_chosen, input_data.Xc, input_data.Xc_ind, input_data.Xc_indptr,
                                                                workspace.buffer_dbl.data(), workspace.buffer_szt.data(), false,
@@ -174,7 +172,7 @@ void split_itree_recursive(std::vector<IsoTree>     &trees,
 
             else
             {
-                if (!workspace.weights_arr.size() && !workspace.weights_map.size())
+                if (!workspace.changed_weights)
                     workspace.this_gain = eval_guided_crit(workspace.ix_arr.data(), workspace.st, workspace.end,
                                                            input_data.categ_data + (workspace.col_chosen - input_data.ncols_numeric) * input_data.nrows,
                                                            input_data.ncat[workspace.col_chosen - input_data.ncols_numeric],
@@ -294,7 +292,7 @@ void split_itree_recursive(std::vector<IsoTree>     &trees,
                 
                 get_split_range(workspace, input_data, model_params, trees.back());
                 if (workspace.unsplittable)
-                    workspace.col_sampler.drop_col(trees.back().col_num + (trees.back().col_type == Numeric)? (size_t)0 : input_data.ncols_numeric);
+                    workspace.col_sampler.drop_col(trees.back().col_num + ((trees.back().col_type == Numeric)? (size_t)0 : input_data.ncols_numeric));
                 else
                     goto produce_split;
             }
@@ -320,7 +318,7 @@ void split_itree_recursive(std::vector<IsoTree>     &trees,
                 get_split_range(workspace, input_data, model_params, trees.back());
                 if (workspace.unsplittable)
                 {
-                    workspace.col_sampler.drop_col(trees.back().col_num + (trees.back().col_type == Numeric)? (size_t)0 : input_data.ncols_numeric);
+                    workspace.col_sampler.drop_col(trees.back().col_num + ((trees.back().col_type == Numeric)? (size_t)0 : input_data.ncols_numeric));
                     workspace.ntried++;
                     if (!workspace.try_all && workspace.ntried >= threshold_shuffle)
                     {
@@ -359,7 +357,7 @@ void split_itree_recursive(std::vector<IsoTree>     &trees,
                 {
                     if (input_data.Xc_indptr == NULL)
                     {
-                        if (!workspace.weights_arr.size() && !workspace.weights_map.size())
+                        if (!workspace.changed_weights)
                             eval_guided_crit(workspace.ix_arr.data(), workspace.st, workspace.end,
                                              input_data.numeric_data + trees.back().col_num * input_data.nrows,
                                              workspace.buffer_dbl.data(), true,
@@ -399,7 +397,7 @@ void split_itree_recursive(std::vector<IsoTree>     &trees,
 
                     else
                     {
-                        if (!workspace.weights_arr.size() && !workspace.weights_map.size())
+                        if (!workspace.changed_weights)
                             eval_guided_crit(workspace.ix_arr.data(), workspace.st, workspace.end,
                                              trees.back().col_num, input_data.Xc, input_data.Xc_ind, input_data.Xc_indptr,
                                              workspace.buffer_dbl.data(), workspace.buffer_szt.data(), true,
@@ -484,7 +482,7 @@ void split_itree_recursive(std::vector<IsoTree>     &trees,
 
                             default:
                             {
-                                if (!workspace.weights_arr.size() && !workspace.weights_map.size())
+                                if (!workspace.changed_weights)
                                     eval_guided_crit(workspace.ix_arr.data(), workspace.st, workspace.end,
                                                      input_data.categ_data + trees.back().col_num * input_data.nrows, input_data.ncat[trees.back().col_num],
                                                      workspace.buffer_szt.data(), workspace.buffer_szt.data() + input_data.max_categ,
@@ -553,7 +551,7 @@ void split_itree_recursive(std::vector<IsoTree>     &trees,
                             default:
                             {
                                 trees.back().cat_split.resize(input_data.ncat[trees.back().col_num]);
-                                if (!workspace.weights_arr.size() && !workspace.weights_map.size())
+                                if (!workspace.changed_weights)
                                     eval_guided_crit(workspace.ix_arr.data(), workspace.st, workspace.end,
                                                      input_data.categ_data + trees.back().col_num * input_data.nrows, input_data.ncat[trees.back().col_num],
                                                      workspace.buffer_szt.data(), workspace.buffer_szt.data() + input_data.max_categ,
@@ -613,9 +611,65 @@ void split_itree_recursive(std::vector<IsoTree>     &trees,
         /* compute statistics for NAs and remember recursion indices/weights */
         if (model_params.missing_action != Fail)
         {
-            trees.back().pct_tree_left = (long double)(workspace.st_NA - workspace.st)
-                                            /
-                                         (long double)(workspace.end - workspace.st + 1 - (workspace.end_NA - workspace.st_NA));
+            if (!workspace.changed_weights)
+            {
+                /* https://sourceforge.net/p/mingw-w64/bugs/909/ */
+                #if !defined(_WIN32) && !defined(_WIN64)
+                trees.back().pct_tree_left = (long double)(workspace.st_NA - workspace.st)
+                                                /
+                                             (long double)(workspace.end - workspace.st + 1 - (workspace.end_NA - workspace.st_NA));
+                #else
+                trees.back().pct_tree_left = (double)(workspace.st_NA - workspace.st)
+                                                /
+                                             (double)(workspace.end - workspace.st + 1 - (workspace.end_NA - workspace.st_NA));
+                #endif
+
+                if (model_params.missing_action == Divide && workspace.st_NA < workspace.end_NA)
+                {
+                    workspace.changed_weights = true;
+
+                    if (input_data.Xc_indptr != NULL && model_params.sample_size < input_data.nrows / 20) {
+                        workspace.weights_arr.clear();
+                        workspace.weights_map.reserve(workspace.end - workspace.st + 1);
+                        for (size_t row = workspace.st; row < workspace.end_NA; row++)
+                            workspace.weights_map[workspace.ix_arr[row]] = 1;
+                    }
+
+                    else {
+                        workspace.weights_arr.resize(input_data.nrows);
+                        for (size_t row = workspace.st; row < workspace.end_NA; row++)
+                            workspace.weights_arr[workspace.ix_arr[row]] = 1;
+                    }
+                }
+            }
+
+            else
+            {
+                /* https://sourceforge.net/p/mingw-w64/bugs/909/ */
+                #if !defined(_WIN32) && !defined(_WIN64)
+                long double sum_weight_left = 0;
+                long double sum_weight_right = 0;
+                #else
+                double sum_weight_left = 0;
+                double sum_weight_right = 0;
+                #endif
+
+                if (workspace.weights_arr.size()) {
+                    for (size_t row = workspace.st; row < workspace.st_NA; row++)
+                        sum_weight_left += workspace.weights_arr[workspace.ix_arr[row]];
+                    for (size_t row = workspace.end_NA; row <= workspace.end; row++)
+                        sum_weight_right += workspace.weights_arr[workspace.ix_arr[row]];
+                }
+
+                else {
+                    for (size_t row = workspace.st; row < workspace.st_NA; row++)
+                        sum_weight_left += workspace.weights_map[workspace.ix_arr[row]];
+                    for (size_t row = workspace.end_NA; row <= workspace.end; row++)
+                        sum_weight_right += workspace.weights_map[workspace.ix_arr[row]];
+                }
+
+                trees.back().pct_tree_left = sum_weight_left / (sum_weight_left + sum_weight_right);
+            }
 
             switch(model_params.missing_action)
             {
@@ -631,23 +685,35 @@ void split_itree_recursive(std::vector<IsoTree>     &trees,
 
                 case Divide:
                 {
-                    if (workspace.weights_map.size())
-                        for (size_t row = workspace.st_NA; row < workspace.end_NA; row++)
-                            workspace.weights_map[workspace.ix_arr[row]] *= trees.back().pct_tree_left;
-                    else
+                    if (workspace.weights_arr.size())
                         for (size_t row = workspace.st_NA; row < workspace.end_NA; row++)
                             workspace.weights_arr[workspace.ix_arr[row]] *= trees.back().pct_tree_left;
+                    else
+                        for (size_t row = workspace.st_NA; row < workspace.end_NA; row++)
+                            workspace.weights_map[workspace.ix_arr[row]] *= trees.back().pct_tree_left;
                     workspace.end = workspace.end_NA - 1;
                     break;
                 }
+
+                default:
+                {
+                    unexpected_error();
+                    break;
+                }
             }
         }
 
         else
         {
+            #if !defined(_WIN32) && !defined(_WIN64)
             trees.back().pct_tree_left = (long double) (workspace.split_ix - workspace.st)
                                             /
                                          (long double) (workspace.end - workspace.st + 1);
+            #else
+            trees.back().pct_tree_left = (double) (workspace.split_ix - workspace.st)
+                                            /
+                                         (double) (workspace.end - workspace.st + 1);
+            #endif
             workspace.end = workspace.split_ix - 1;
         }
 
@@ -694,15 +760,36 @@ void split_itree_recursive(std::vector<IsoTree>     &trees,
 
                 case Divide:
                 {
-                    if (workspace.weights_map.size())
+                    if (!workspace.changed_weights && workspace.st_NA < workspace.end_NA)
+                    {
+                        workspace.changed_weights = true;
+
+                        if (workspace.weights_arr.size()) {
+                            for (size_t row = workspace.st_NA; row <= workspace.end; row++)
+                                workspace.weights_arr[workspace.ix_arr[row]] = 1;
+                        }
+
+                        else {
+                            for (size_t row = workspace.st_NA; row <= workspace.end; row++)
+                                workspace.weights_map[workspace.ix_arr[row]] = 1;
+                        }
+                    }
+
+                    if (workspace.weights_arr.size())
                         for (size_t row = workspace.st_NA; row < workspace.end_NA; row++)
-                            workspace.weights_map[workspace.ix_arr[row]] *= (1 - trees[tree_from].pct_tree_left);
+                            workspace.weights_arr[workspace.ix_arr[row]] *= (1. - trees[tree_from].pct_tree_left);
                     else
                         for (size_t row = workspace.st_NA; row < workspace.end_NA; row++)
-                            workspace.weights_arr[workspace.ix_arr[row]] *= (1 - trees[tree_from].pct_tree_left);
+                            workspace.weights_map[workspace.ix_arr[row]] *= (1. - trees[tree_from].pct_tree_left);
                     workspace.st = workspace.st_NA;
                     break;
                 }
+
+                default:
+                {
+                    unexpected_error();
+                    break;
+                }
             }
         }
 
@@ -726,14 +813,14 @@ void split_itree_recursive(std::vector<IsoTree>     &trees,
     /* if it reached the limit, calculate terminal statistics */
     terminal_statistics:
     {
-        if (!workspace.weights_arr.size() && !workspace.weights_map.size())
+        if (!workspace.changed_weights)
         {
             trees.back().score = (double)(curr_depth + expected_avg_depth(workspace.end - workspace.st + 1));
         }
 
         else
         {
-            if (sum_weight == -HUGE_VAL)
+            if (sum_weight <= -HUGE_VAL)
                 sum_weight = calculate_sum_weights(workspace.ix_arr, workspace.st, workspace.end, curr_depth,
                                                    workspace.weights_arr, workspace.weights_map);
             trees.back().score = (double)(curr_depth + expected_avg_depth(sum_weight));
@@ -742,10 +829,8 @@ void split_itree_recursive(std::vector<IsoTree>     &trees,
         trees.back().cat_split.clear();
         trees.back().cat_split.shrink_to_fit();
 
-        trees.back().remainder = workspace.weights_arr.size()?
-                                 sum_weight : (workspace.weights_map.size()?
-                                               sum_weight : ((double)(workspace.end - workspace.st + 1))
-                                               );
+        trees.back().remainder = workspace.changed_weights?
+                                    (double)sum_weight : (double)(workspace.end - workspace.st + 1);
 
         /* for distance, assume also the elements keep being split */
         if (model_params.calc_dist)
@@ -754,7 +839,7 @@ void split_itree_recursive(std::vector<IsoTree>     &trees,
         /* add this depth right away if requested */
         if (workspace.row_depths.size())
         {
-            if (!workspace.weights_arr.size() && !workspace.weights_map.size())
+            if (!workspace.changed_weights)
             {
                 for (size_t row = workspace.st; row <= workspace.end; row++)
                     workspace.row_depths[workspace.ix_arr[row]] += trees.back().score;
