@@ -47,10 +47,10 @@
 #define ISOTREE_H
 
 /* Standard headers */
-#include <stddef.h>
-#include <math.h>
-#include <limits.h>
-#include <string.h>
+#include <cstddef>
+#include <cmath>
+#include <climits>
+#include <cstring>
 #include <vector>
 #include <iterator>
 #include <numeric>
@@ -61,9 +61,14 @@
 #include <memory>
 #include <utility>
 #include <cstdint>
+#include <exception>
+#include <stdexcept>
+#include <cassert>
 #include <iostream>
 #ifndef _FOR_R
-    #include <stdio.h> 
+    #include <cstdio>
+    using std::printf;
+    using std::fprintf;
 #else
     extern "C" {
         #include <R_ext/Print.h>
@@ -84,9 +89,16 @@
 #ifdef _FOR_R
     #include <Rcpp.h>
 #endif
-#include <signal.h>
+#include <csignal>
 typedef void (*sig_t_)(int);
+using std::signal;
+using std::raise;
 
+using std::size_t;
+using std::memset;
+using std::memcpy;
+
+#define unexpected_error() throw std::runtime_error("Unexpected error. Please open an issue in GitHub.\n")
 
 /* By default, will use Mersenne-Twister for RNG, but can be switched to something faster */
 #ifdef _USE_MERSENNE_TWISTER
@@ -113,20 +125,28 @@ typedef void (*sig_t_)(int);
 #define ix_parent(ix) (((ix) - 1) / 2)  /* integer division takes care of deciding left-right */
 #define ix_child(ix)  (2 * (ix) + 1)
 /* https://stackoverflow.com/questions/101439/the-most-efficient-way-to-implement-an-integer-based-power-function-powint-int */
-#if defined(__LITTLE_ENDIAN) && defined(__BYTE_ORDER) && (IS_LITTLE_ENDIAN == 1)
+#if defined(__LITTLE_ENDIAN) && defined(__BYTE_ORDER) && (__BYTE_ORDER == __LITTLE_ENDIAN)
     #define pow2(n) ( ((size_t) 1) << (n) )
 #else
     #define pow2(n) ((size_t)powl((long double)2, (long double)n))
 #endif
 #define square(x) ((x) * (x))
+#if defined(__GNUC__) && (__GNUC__ >= 5)
+    #pragma GCC diagnostic push
+    #pragma GCC diagnostic ignored "-Wmaybe-uninitialized"
+#elif defined(__clang__) && !defined(_FOR_R)
+    #pragma clang diagnostic push
+    #pragma clang diagnostic ignored "-Wuninitialized"
+#endif
 /* https://stackoverflow.com/questions/2249731/how-do-i-get-bit-by-bit-data-from-an-integer-value-in-c */
 #define extract_bit(number, bit) (((number) >> (bit)) & 1)
-#ifndef isinf
-    #define isinf std::isinf
-#endif
-#ifndef isnan
-    #define isnan std::isnan
+#if defined(__GNUC__) && (__GNUC__ >= 5)
+    #pragma GCC diagnostic pop
+#elif defined(__clang__) && !defined(_FOR_R)
+    #pragma clang diagnostic pop
 #endif
+using std::isinf;
+using std::isnan;
 #define is_na_or_inf(x) (isnan(x) || isinf(x))
 
 
@@ -140,7 +160,7 @@ typedef void (*sig_t_)(int);
 /* MSVC is stuck with an OpenMP version that's 19 years old at the time of writing and does not support unsigned iterators */
 #ifdef _OPENMP
     #if (_OPENMP < 200801) || defined(_WIN32) || defined(_WIN64) /* OpenMP < 3.0 */
-        #define size_t_for long
+        #define size_t_for long long
     #else
         #define size_t_for size_t
     #endif
@@ -148,6 +168,16 @@ typedef void (*sig_t_)(int);
     #define size_t_for size_t
 #endif
 
+#if defined(_FOR_R) || defined(_FOR_PYTHON) || !defined(_WIN32)
+    #define ISOTREE_EXPORTED 
+#else
+    #ifdef ISOTREE_COMPILE_TME
+        #define ISOTREE_EXPORTED __declspec(dllexport)
+    #else
+        #define ISOTREE_EXPORTED __declspec(dllimport)
+    #endif 
+#endif
+
 
 /*   Apple at some point decided to drop OMP library and headers from its compiler distribution
 *    and to alias 'gcc' to 'clang', which work differently when given flags they cannot interpret,
@@ -478,7 +508,7 @@ struct ImputedData {
     ImputedData(InputData &input_data, size_t row)
     {
         initialize_impute_calc(*this, input_data, row);
-    };
+    }
 
 };
 
@@ -531,17 +561,18 @@ struct WorkerMemory {
     size_t               end_NA;
     size_t               split_ix;
     std::unordered_map<size_t, double> weights_map;
-    std::vector<double>  weights_arr;    /* when not ignoring NAs and when using weights as density */
+    std::vector<double>  weights_arr;     /* when not ignoring NAs and when using weights as density */
+    bool                 changed_weights; /* when using 'missing_action'='Divide' or density weights */
     double               xmin;
     double               xmax;
-    size_t               npresent;       /* 'npresent' and 'ncols_tried' are used interchangeable and for unrelated things */
+    size_t               npresent;        /* 'npresent' and 'ncols_tried' are used interchangeable and for unrelated things */
     bool                 unsplittable;
     std::vector<bool>    is_repeated;
     std::vector<char>    categs;
-    size_t               ncols_tried;    /* 'npresent' and 'ncols_tried' are used interchangeable and for unrelated things */
+    size_t               ncols_tried;     /* 'npresent' and 'ncols_tried' are used interchangeable and for unrelated things */
     int                  ncat_tried;
-    std::vector<double>  btree_weights;  /* only when using weights for sampling */
-    ColumnSampler        col_sampler;    /* columns can get eliminated, keep a copy for each thread */
+    std::vector<double>  btree_weights;   /* only when using weights for sampling */
+    ColumnSampler        col_sampler;     /* columns can get eliminated, keep a copy for each thread */
 
     /* for split criterion */
     std::vector<double>  buffer_dbl;
@@ -602,6 +633,15 @@ typedef struct WorkerForSimilarity {
     bool                assume_full_distr; /* doesn't need to have one copy per worker */
 } WorkerForSimilarity;
 
+typedef struct WorkerForPredictCSC {
+    std::vector<size_t> ix_arr;
+    size_t              st;
+    size_t              end;
+    std::vector<double> comb_val;
+    std::vector<double> weights_arr;
+    std::vector<double> depths;
+} WorkerForPredictCSC;
+
 class RecursionState {
 public:
     size_t  st;
@@ -611,6 +651,7 @@ public:
     size_t  end;
     size_t  sampler_pos;
     size_t  n_dropped;
+    bool    changed_weights;
     bool    full_state;
     std::vector<size_t> ix_arr;
     std::vector<bool>   cols_possible;
@@ -662,7 +703,7 @@ int add_tree(IsoForest *model_outputs, ExtIsoForest *model_outputs_ext,
              double min_gain, MissingAction missing_action,
              CategSplit cat_split_type, NewCategAction new_cat_action,
              UseDepthImp depth_imp, WeighImpRows weigh_imp_rows,
-             bool   all_perm, std::vector<ImputeNode> *impute_nodes, size_t min_imp_obs,
+             bool   all_perm, Imputer *imputer, size_t min_imp_obs,
              uint64_t random_seed);
 template <class InputData, class WorkerMemory>
 void fit_itree(std::vector<IsoTree>    *tree_root,
@@ -740,6 +781,33 @@ void traverse_hplane(std::vector<IsoHPlane>   &hplane,
                      ImputedData             *imputed_data,
                      sparse_ix *restrict      tree_num,
                      size_t                   row);
+template <class real_t, class sparse_ix>
+void batched_csc_predict(PredictionData<real_t, sparse_ix> &prediction_data, int nthreads,
+                         IsoForest *model_outputs, ExtIsoForest *model_outputs_ext,
+                         double output_depths[],   sparse_ix tree_num[]);
+template <class PredictionData, class sparse_ix>
+void traverse_itree_csc(WorkerForPredictCSC   &workspace,
+                        std::vector<IsoTree>  &trees,
+                        IsoForest             &model_outputs,
+                        PredictionData        &prediction_data,
+                        sparse_ix             *tree_num,
+                        size_t                curr_tree,
+                        bool                  has_range_penalty);
+template <class PredictionData, class sparse_ix>
+void traverse_hplane_csc(WorkerForPredictCSC      &workspace,
+                         std::vector<IsoHPlane>   &hplanes,
+                         ExtIsoForest             &model_outputs,
+                         PredictionData           &prediction_data,
+                         sparse_ix                *tree_num,
+                         size_t                   curr_tree,
+                         bool                     has_range_penalty);
+template <class PredictionData>
+void add_csc_range_penalty(WorkerForPredictCSC  &workspace,
+                           PredictionData       &prediction_data,
+                           double               *weights_arr,
+                           size_t               col_num,
+                           double               range_low,
+                           double               range_high);
 template <class PredictionData>
 double extract_spC(PredictionData &prediction_data, size_t row, size_t col_num);
 template <class PredictionData, class sparse_ix>
@@ -932,8 +1000,13 @@ void get_range(size_t ix_arr[], size_t st, size_t end, size_t col_num,
                MissingAction missing_action, double &xmin, double &xmax, bool &unsplittable);
 void get_categs(size_t ix_arr[], int x[], size_t st, size_t end, int ncat,
                 MissingAction missing_action, char categs[], size_t &npresent, bool &unsplittable);
+#if !defined(_WIN32) && !defined(_WIN64)
 long double calculate_sum_weights(std::vector<size_t> &ix_arr, size_t st, size_t end, size_t curr_depth,
                                   std::vector<double> &weights_arr, std::unordered_map<size_t, double> &weights_map);
+#else
+     double calculate_sum_weights(std::vector<size_t> &ix_arr, size_t st, size_t end, size_t curr_depth,
+                                  std::vector<double> &weights_arr, std::unordered_map<size_t, double> &weights_map);
+#endif
 extern bool interrupt_switch;
 extern bool signal_is_locked;
 void set_interrup_global_variable(int s);
@@ -1130,39 +1203,40 @@ double eval_guided_crit_weighted(size_t *restrict ix_arr, size_t st, size_t end,
                                  mapping w);
 
 /* merge_models.cpp */
+ISOTREE_EXPORTED
 void merge_models(IsoForest*     model,      IsoForest*     other,
                   ExtIsoForest*  ext_model,  ExtIsoForest*  ext_other,
                   Imputer*       imputer,    Imputer*       iother);
 
 #if defined(_ENABLE_CEREAL) || defined(_FOR_PYTHON)
 /* serialize.cpp */
-void serialize_isoforest(IsoForest &model, std::ostream &output);
-void serialize_isoforest(IsoForest &model, const char *output_file_path);
-std::string serialize_isoforest(IsoForest &model);
-void deserialize_isoforest(IsoForest &output_obj, std::istream &serialized);
-void deserialize_isoforest(IsoForest &output_obj, const char *input_file_path);
-void deserialize_isoforest(IsoForest &output_obj, std::string &serialized, bool move_str);
-void serialize_ext_isoforest(ExtIsoForest &model, std::ostream &output);
-void serialize_ext_isoforest(ExtIsoForest &model, const char *output_file_path);
-std::string serialize_ext_isoforest(ExtIsoForest &model);
-void deserialize_ext_isoforest(ExtIsoForest &output_obj, std::istream &serialized);
-void deserialize_ext_isoforest(ExtIsoForest &output_obj, const char *input_file_path);
-void deserialize_ext_isoforest(ExtIsoForest &output_obj, std::string &serialized, bool move_str);
-void serialize_imputer(Imputer &imputer, std::ostream &output);
-void serialize_imputer(Imputer &imputer, const char *output_file_path);
-std::string serialize_imputer(Imputer &imputer);
-void deserialize_imputer(Imputer &output_obj, std::istream &serialized);
-void deserialize_imputer(Imputer &output_obj, const char *input_file_path);
-void deserialize_imputer(Imputer &output_obj, std::string &serialized, bool move_str);
+ISOTREE_EXPORTED void serialize_isoforest(IsoForest &model, std::ostream &output);
+ISOTREE_EXPORTED void serialize_isoforest(IsoForest &model, const char *output_file_path);
+ISOTREE_EXPORTED std::string serialize_isoforest(IsoForest &model);
+ISOTREE_EXPORTED void deserialize_isoforest(IsoForest &output_obj, std::istream &serialized);
+ISOTREE_EXPORTED void deserialize_isoforest(IsoForest &output_obj, const char *input_file_path);
+ISOTREE_EXPORTED void deserialize_isoforest(IsoForest &output_obj, std::string &serialized, bool move_str);
+ISOTREE_EXPORTED void serialize_ext_isoforest(ExtIsoForest &model, std::ostream &output);
+ISOTREE_EXPORTED void serialize_ext_isoforest(ExtIsoForest &model, const char *output_file_path);
+ISOTREE_EXPORTED std::string serialize_ext_isoforest(ExtIsoForest &model);
+ISOTREE_EXPORTED void deserialize_ext_isoforest(ExtIsoForest &output_obj, std::istream &serialized);
+ISOTREE_EXPORTED void deserialize_ext_isoforest(ExtIsoForest &output_obj, const char *input_file_path);
+ISOTREE_EXPORTED void deserialize_ext_isoforest(ExtIsoForest &output_obj, std::string &serialized, bool move_str);
+ISOTREE_EXPORTED void serialize_imputer(Imputer &imputer, std::ostream &output);
+ISOTREE_EXPORTED void serialize_imputer(Imputer &imputer, const char *output_file_path);
+ISOTREE_EXPORTED std::string serialize_imputer(Imputer &imputer);
+ISOTREE_EXPORTED void deserialize_imputer(Imputer &output_obj, std::istream &serialized);
+ISOTREE_EXPORTED void deserialize_imputer(Imputer &output_obj, const char *input_file_path);
+ISOTREE_EXPORTED void deserialize_imputer(Imputer &output_obj, std::string &serialized, bool move_str);
 #ifdef _MSC_VER
-void serialize_isoforest(IsoForest &model, const wchar_t *output_file_path);
-void deserialize_isoforest(IsoForest &output_obj, const wchar_t *input_file_path);
-void serialize_ext_isoforest(ExtIsoForest &model, const wchar_t *output_file_path);
-void deserialize_ext_isoforest(ExtIsoForest &output_obj, const wchar_t *input_file_path);
-void serialize_imputer(Imputer &imputer, const wchar_t *output_file_path);
-void deserialize_imputer(Imputer &output_obj, const wchar_t *input_file_path);
+ISOTREE_EXPORTED void serialize_isoforest(IsoForest &model, const wchar_t *output_file_path);
+ISOTREE_EXPORTED void deserialize_isoforest(IsoForest &output_obj, const wchar_t *input_file_path);
+ISOTREE_EXPORTED void serialize_ext_isoforest(ExtIsoForest &model, const wchar_t *output_file_path);
+ISOTREE_EXPORTED void deserialize_ext_isoforest(ExtIsoForest &output_obj, const wchar_t *input_file_path);
+ISOTREE_EXPORTED void serialize_imputer(Imputer &imputer, const wchar_t *output_file_path);
+ISOTREE_EXPORTED void deserialize_imputer(Imputer &output_obj, const wchar_t *input_file_path);
 #endif /* _MSC_VER */
-bool has_msvc();
+ISOTREE_EXPORTED bool has_msvc();
 #ifdef _FOR_PYTHON
 void serialize_isoforest(IsoForest &model, void *output_file_path);
 void deserialize_isoforest(IsoForest &output_obj, void *input_file_path);
@@ -1176,11 +1250,13 @@ bool has_cereal();
 #endif /* _ENABLE_CEREAL || _FOR_PYTON */
 
 /* sql.cpp */
+ISOTREE_EXPORTED
 std::vector<std::string> generate_sql(IsoForest *model_outputs, ExtIsoForest *model_outputs_ext,
                                       std::vector<std::string> &numeric_colnames, std::vector<std::string> &categ_colnames,
                                       std::vector<std::vector<std::string>> &categ_levels,
                                       bool output_tree_num, bool index1, bool single_tree, size_t tree_num,
                                       int nthreads);
+ISOTREE_EXPORTED
 std::string generate_sql_with_select_from(IsoForest *model_outputs, ExtIsoForest *model_outputs_ext,
                                           std::string &table_from, std::string &select_as,
                                           std::vector<std::string> &numeric_colnames, std::vector<std::string> &categ_colnames,
@@ -1198,9 +1274,4 @@ void extract_cond_ext_isotree(ExtIsoForest &model, IsoHPlane &hplane,
                               std::vector<std::string> &numeric_colnames, std::vector<std::string> &categ_colnames,
                               std::vector<std::vector<std::string>> &categ_levels);
 
-/* dealloc.cpp */
-void dealloc_IsoForest(IsoForest &model_outputs);
-void dealloc_IsoExtForest(ExtIsoForest &model_outputs_ext);
-void dealloc_Imputer(Imputer &imputer);
-
 #endif /* ISOTREE_H */
